/// ECorrSimulator.cc -*- c++ -*- Hans-Peter Stoerr Mon Nov 17 2003
/// We try to find the best correction algorithm
/// $Id: ECorrSimulator.cc,v 1.4 2003/12/29 18:06:31 stoerr Exp $

#include <iostream>
#include <stdlib.h>
#include <sys/time.h>

#define P(X) do { std::cerr << #X"=" << X << std::endl; std::cerr.flush(); } while(0)

const int gMaxErrors = 4;

class ECorrSimulator
{
public:
  int numBlocks;
  int numErrors;
  bool *pBlockFaults; // Blocks where errors are

  ECorrSimulator(const int blocks, int errors)
  {
    numBlocks = blocks;
    numErrors = errors;
    pBlockFaults = new bool[blocks];
    int i;
    for (i=0; i<blocks; ++i) pBlockFaults[i]=false;
    while ( errors > 0 ) {
      int rblk = random()%blocks;
      if (!pBlockFaults[rblk]) {
        pBlockFaults[rblk]=true;
        --errors;
      }
    }
  }
  
  ~ECorrSimulator() { delete pBlockFaults; }

  bool hasErrors() { return numErrors > 0; }

  // runs simulation of errorcorrection with eccs for stripe 
  // with stripeSkip
  // returns true if we have found corrected errors
  bool stripeCorrect(const int stripeSkip, const int stripeLen)
  {
    bool haveCorrected=false;   // whether we did correct sth.
    int pos=0, next_pos=-1, checkedBlocks=0, i;
    while ( checkedBlocks < numBlocks ) {
      next_pos=pos;
      int error_count=0;        // number of errors in current stripe
      for (i=0; i<stripeLen; ++i) {
        if (pBlockFaults[next_pos]) ++error_count;
        next_pos = (next_pos + stripeSkip) % numBlocks;
        ++checkedBlocks;
      }
      if ( (0<error_count) && (error_count<gMaxErrors) ) {
        // we have errors to correct
        numErrors -= error_count;
        next_pos=pos;
        for (i=0; i<stripeLen; ++i) { // run through blocks again
          pBlockFaults[next_pos] = false;
          next_pos = (next_pos + stripeSkip) % numBlocks;
        }
        haveCorrected = true;
      }
      pos=next_pos;
    }
    return haveCorrected;
  }

  // runs simulation of errorcorrection with double stripeCorrect
  // returns true if we have found corrected errors
  void doubleStripeCorrect(const int stripeLen) {
    bool changes=true;
    while (changes) {
      changes = stripeCorrect(5,stripeLen);
      changes = changes || stripeCorrect(7,stripeLen);
    }
  }
};

double probCorr(const int numBlocks, const int numErrors, const int stripeLen)
{
  const int n=numBlocks/stripeLen; 
  P(exp(-n*pow(numErrors*1.0/n,3.8)/45)); // gute Approximation.
  int i, cntErr=0;
  for (i=0; i<1000; ++i) {
    ECorrSimulator sim(numBlocks, numErrors);
    sim.stripeCorrect(3,stripeLen);
    if (sim.hasErrors()) ++cntErr;
  }
  return 1-double(cntErr)/i;
}

double probDoubleCorr(const int numBlocks, const int numErrors, const int stripeLen)
{
  int i, cntErr=0;
  for (i=0; i<1000; ++i) {
    ECorrSimulator sim(numBlocks, numErrors);
    sim.doubleStripeCorrect(stripeLen);
    if (sim.hasErrors()) ++cntErr;
  }
  return 1-double(cntErr)/i;
}

int main(/* int argc, char * const argv[] */)
{
  struct timeval tv; gettimeofday(&tv,NULL);
  srandom(tv.tv_sec-tv.tv_usec);

  P(probCorr(100000,320,100));
  P(probCorr(100000,320,249));
  P(probCorr(10000,50,100));
  P(probDoubleCorr(10000,50,100));
  P(probDoubleCorr(10000,50,200));
  P(probCorr(20000,100,100));
  P(probCorr(10000,40,200));
  P(probCorr(1250000,100,100));
  P(probCorr(1250000,200,100));
  P(probCorr(1250000,500,100));
  P(probCorr(1250000,1000,100));
  P(probCorr(1250000,2000,100));
  P(probCorr(1250000,3000,100));
  P(probCorr(1250000,4000,100));
  P(probCorr(1250000,5000,100));
  P(probCorr(1250000,10000,100));
  P(probCorr(1250000,20000,100));
  int num_blocks=100000;
  int hysiz=100;
  double num_errs=40;
  double prob=1;
  while (prob>0.1) {
    prob = probCorr(num_blocks, int(num_errs), hysiz);
    std::cout << int(num_errs) << "\t" << 1-prob << "\n";
    std::cout.flush();
    num_errs = num_errs*pow(2.0,1.0/3);
  }
  return 0;
}
