/// hyperblkcodec.cc -*- c++ -*- Hans-Peter Stoerr Fri Nov  7 2003
/// Protect a number of blocks with a number of error protection information
/// blocks.
/// Protection is in stripes: the nth GF256 of the protected blocks are used
/// to generate the protection information stored in the nth GF256 of the
/// protection information.
/// $Id: hyperblkcodec.cc,v 1.2 2003/12/29 18:06:31 stoerr Exp $

#include "hyperblkcodec.h"

#include <gf256.h>
#include <rsff-f9.h>

const int gfblklen=blklen*sizeof(char)/sizeof(GF256);

/// write error protection info 
/// blks: blocks to protect, hblen*blklen bytes
/// protblks: protection info, hbcorr*blklen bytes
void hyperblockprotect(const char* blks, char* protblks)
{
  const GF256* gblks=(const GF256*) blks;
  GF256* gprot=(GF256*) protblks;

  GF256 data[249], parity[6];
  int i,n;
  for (i=0; i<249; ++i) data[i]=0;
  for (i=0; i<6; ++i) parity[i]=0;

  for (n=0; n<gfblklen; ++n) { // for each stripe
    for (i=0; i<hblen; ++i) data[i] = gblks[n+gfblklen*i];
    rs255_249_encode(data,parity);
    // if (n==0) P(data);
    // if (n==0) P(parity);    
    for (i=0; i<hbcorr; ++i) gprot[n+gfblklen*i] = parity[i];
  }
}

/// sets dst from src iff dst is not equal to src.
/// So we avoid touching memory segments if they are really unmodified.
template <class T> void setifchanged(T& dst, const T& src)
{
  if (src != dst) dst = src;
}

/// reconstruct blocks with error protection info 
/// blks: blocks to correct, hblen*blklen bytes
/// protblks: protection info, hbcorr*blklen bytes
/// corrects blks; returns true iff block could be corrected.
bool hyperblockcorrect(char* blks, const char* protblks)
{
  GF256* gblks=(GF256*) blks;
  const GF256* gprot=(const GF256*) protblks;

  GF256 received[255];
  int i,n, maxerrcorrected=0;
  for (i=0; i<255; ++i) received[i]=0;

  for (n=0; n<gfblklen; ++n) { // for each stripe
    for (i=0; i<hblen; ++i) received[i] = gblks[n+gfblklen*i];
    for (i=0; i<hbcorr; ++i) received[i+249] = gprot[n+gfblklen*i];
    // if (n==0) P(received);
    int status = rs255_249_decode(received);
    if (RS255_249_E4 == status) return false;
    int errcorrected = rs255_249_number_of_errors(status);
    if ( 0 < errcorrected ) {
      for (i=0; i<hblen; ++i) setifchanged(gblks[n+gfblklen*i],received[i]);
      if ( maxerrcorrected < errcorrected ) maxerrcorrected = errcorrected;
    }
  }
  return true;
}

/// testroutine. Set blklen in options.h to 16 or so.
void checkErrorProtection()
{
  char in[hblen*blklen+1]; in[hblen*blklen]=0;
  char prot[hbcorr*blklen+1]; prot[hbcorr*blklen]=0;
  int i;
  for (i=0; i<hblen*blklen; ++i) in[i]='a'+(i%23);
  for (i=0; i<hbcorr*blklen; ++i) prot[i]='z';
  P(in);
  // P(prot);
  hyperblockprotect(in,prot);
  // P(in);
  P(prot);
  in[3]='z'; in[22]='y';
  P(in);
  P(hyperblockcorrect(in,prot));
  P(in);
  // P(prot);  
}

