/// protectcorrectfile.cc -*- c++ -*- Hans-Peter Stoerr Wed Nov 12 2003
/// $Id: protectcorrectfile.cc,v 1.13 2003/12/30 11:05:15 stoerr Exp $

#include "protectcorrectfile.h"

#include "options.h"
#include "header.h"
#include "footprint.h"
#include "hyperblkcodec.h"

#include <sys/stat.h>
#include <assert.h>


/// Calculate distance to next block for file striping
/// Filelength len (in blocks), at least stripenum stripes
long interleaveskip(const long len, const long stripenum)
{
  long sskip = len/stripenum+1;
  // decrease sskip until the cycle modulo len is len.
  long i,s;
  do {
    --sskip;
    i=0; s=0;
    do { ++i; s = (s+sskip)%len; } while (0!=s);
    // P(sskip); P(i);
  } while(len != i);
  return sskip;
}

/// read a file fil into buffer buf while interleaving it for better correction.
/// returns buffer to read into
/// fblocks: number of blklen char blocks in infile
/// stripe: striping parameter
char* readandinterleave(int fblocks, int stripe, FILE* infile)
{
  char* buf = (char*) malloc((fblocks+hblen)*blklen*sizeof(char));
  // we avoid using new so the (very large!) buffer is not zeroed.
  if (NULL == buf) {
    std::cerr << progname << ": can't allocate enough memory ("
              << (fblocks+hblen)*blklen*sizeof(char) << " bytes)\n";
    abort();
  }
  // but we zero the last few blocks since they might be used later.
  memset(buf+(fblocks)*blklen,0,hblen*blklen*sizeof(char));
  long blkpos=0;
  for (long blkcount=0; blkcount<fblocks; ++blkcount) {
    if ( 0 == fread(buf+(blkpos*blklen), sizeof(char), blklen, infile) )
      { perror("readandinterleave"); abort(); }
    blkpos = (blkpos+stripe)%fblocks;
  }
  return buf;
}

/// write a file from interleaved buffer. Inverse of readandinterleave.
/// fblocks: number of blklen char blocks in infile
/// stripe: striping parameter
/// outfile: file to write to
void writeanddeinterleave(char* buf, int fblocks, int stripe, FILE* outfile)
{
  long blkpos=0;
  for (long blkcount=0; blkcount<fblocks; ++blkcount) {
    if ( 0 == fwrite(buf+(blkpos*blklen), sizeof(char)*blklen, 1, outfile) )
      { perror("writeanddeinterleave"); abort(); }
    blkpos = (blkpos+stripe)%fblocks;
  }
}

/// protect the file protectedfile
void protectfile()
{
  long i;
  long fsize;                   // file size of protectedfile
  long fblocks;                 // file size in blocks
  long stripeskip;              // striping parameter
  long ostripeskip;             // eccfile striping parameter
  FILE* infile=fopen(protectedfname,"rb");
  if (NULL == infile) { perror(protectedfname); abort(); }
  // calculate file size
  if (0 != fseek(infile, 0, SEEK_END) ) { perror(protectedfname); abort(); }
  fsize = ftell(infile);  P(fsize);
  fseek(infile,0,SEEK_SET);
  fblocks = fsize/blklen;
  if (fblocks*blklen<fsize) ++fblocks; // round up
  P(fblocks); assert(fblocks*blklen>=fsize);
  stripeskip=interleaveskip(fblocks,hblen);  P(stripeskip);
  
  // open eccfile
  FILE* eccfile=fopen(eccfname,"wb");
  if (NULL == eccfile) { perror(eccfname); abort(); }

  // now do the interleaving.
  if (verbose) {
    std::cout << "Reading " << protectedfname << std::endl;
    std::cout.flush();
  }
  char* inbuf = readandinterleave(fblocks,stripeskip,infile);
  assert(ftell(infile)==fsize); // P(inbuf);

  addheaderitem("protectedfile",protectedfname);
  addheaderitem("fsize",fsize);
  {
    char stim[64]; struct stat buf;
    stat(protectedfname,&buf);
    strftime(stim,64,"%Y-%m-%d %T",localtime(&buf.st_mtime));
    addheaderitem("fdate",stim);
  }
  addheaderitem("footprint",footprint(infile));
  {
    char stim[64]; time_t tim; time(&tim);
    strftime(stim,64,"%Y-%m-%d %T",localtime(&tim));
    addheaderitem("date",stim);
  }
  addheaderitem("blklen",blklen);
  addheaderitem("hblen",hblen);
  addheaderitem("hbcorr",hbcorr);
  addheaderitem("stripeskip",stripeskip);
  fclose(infile);

  // calculate ecc data
  char *protbuf, *writeptr;
  const long eccblocks=hbcorr*long(ceil(double(fblocks)/hblen));
  P(eccblocks);
  // we avoid using new so the (very large!) buffer is not zeroed.
  protbuf = (char*) malloc(eccblocks*blklen*sizeof(char));
  writeptr=protbuf;
  // for(i=0; i<eccblocks*blklen; ++i) protbuf[i]='y'; // for testing
  if (verbose) {
    std::cout << "Calculating ... " << std::endl; std::cout.flush();
  }
  for(i=0; i<fblocks ; i+=hblen) {
    hyperblockprotect(inbuf+i*blklen,writeptr);
    writeptr += hbcorr*blklen;
  }
  assert((writeptr-protbuf)/blklen == eccblocks);
  // P(protbuf); // P(crc(protbuf,eccblocks*blklen*sizeof(char)));
  // write ecc data to file
  ostripeskip=interleaveskip(eccblocks,hbcorr);  P(ostripeskip);
  addheaderitem("ostripeskip",ostripeskip);
  writeheader(eccfile); // first copy of header
  if (verbose) {
    std::cout << "Writing " << eccfname << std::endl; std::cout.flush();
  }
  writeanddeinterleave(protbuf, eccblocks, ostripeskip, eccfile);
  writeheader(eccfile); // second copy of header
  fclose(eccfile);
  free(inbuf);
}

/// correct the protectedfile to correctedfile
bool correctfile()
{
  long protfsize;                   // file size of protectedfile
  long protfblocks;                 // file size in blocks
  long eccfsize;                   // file size of eccfile
  long eccblocks;      // number of ecc hyperblocks (hbcorr blks) in eccfile
  long stripeskip;              // striping parameter
  long ostripeskip;              // striping parameter for ecc file
  // open protected file (to correct)
  FILE* protectedfile=fopen(protectedfname,"rb");
  if (NULL == protectedfile) { perror(protectedfname); abort(); }
  // calculate file size
  if (0 != fseek(protectedfile, 0, SEEK_END) ) { perror(protectedfname); abort(); }
  protfsize = ftell(protectedfile); rewind(protectedfile);
  P(protfsize);
  protfblocks = protfsize/blklen;
  if (protfblocks*blklen<protfsize) ++protfblocks; // round up
  P(protfblocks); assert(protfblocks*blklen>=protfsize);
  // open ecc file 
  FILE* eccfile=fopen(eccfname,"rb");
  if (NULL == eccfile) { perror(protectedfname); abort(); }
  // calculate file size
  if (0 != fseek(eccfile, 0, SEEK_END) ) { perror(protectedfname); abort(); }
  eccfsize = ftell(eccfile);  P(eccfsize);
  fseek(eccfile,0,SEEK_SET);
  eccblocks = (eccfsize-2*headerlen)/blklen;
  P(eccblocks); assert(eccblocks*blklen+2*headerlen == eccfsize);
  // read header
  char* header=readheader(eccfile); // P(header);
  while(*header) {
    char* value=index(header,'='); *(value++)=0;
    char* key=header;
    header=index(value,';'); *(header++)=0;
    if ( 0 == strcmp(key,"protectedfile") ) {
      P(key);
      P(/*protectedfile*/value);
      // ignore for now
    } else if ( 0 == strcmp(key,"fsize") ) {
      P(key);
      P(value);
      assert ( atol(value) == protfsize );
      // XXX: we should ask what to do, here.
    } else if ( 0 == strcmp(key,"fdate") ) {
      P(key);
      P(/*fdate*/value);
      // ignore for now
    } else if ( 0 == strcmp(key,"date") ) {
      P(key);
      P(/*date*/value);
      // ignore for now
    } else if ( 0 == strcmp(key,"footprint") ) {
      char *protfprint = footprint(protectedfile);
      P(protfprint);
      P(     value);
      if (0 != strcmp(protfprint, value) ) {
        std::cerr << protectedfname << ": footprints are not conforming ("
                  << int(100*footPrintDiff(protfprint, value)) << "%) :"
                  << protfprint << " versus " << value << "\n";
      }
      // XXX: we should ask whether we should continue.
    } else if ( 0 == strcmp(key,"blklen") ) {
      // blklen=atol(value);
      assert( blklen == atol(value) ); // const
      P(blklen);
    } else if ( 0 == strcmp(key,"hblen") ) {
      hblen=atol(value); P(hblen); 
    } else if ( 0 == strcmp(key,"hbcorr") ) {
      // hbcorr=atol(value);
      assert( hbcorr == atol(value) );  // const
      P(hbcorr);
    } else if ( 0 == strcmp(key,"stripeskip") ) {
      stripeskip=atol(value); P(stripeskip);
    } else if ( 0 == strcmp(key,"ostripeskip") ) {
      ostripeskip=atol(value); P(ostripeskip);
    } else {
      std::cerr << progname << ": unknown tag " << key << "=" << value
                << " in eccfile " << eccfname << "\n";
    }
  }
  
  // read and interleave the protectedfile
  rewind(protectedfile);
  if (verbose) {
    std::cout << "Reading " << protectedfname << std::endl; std::cout.flush();
  }
  char* inbuf=readandinterleave(protfblocks,stripeskip,protectedfile);
  assert(ftell(protectedfile)==protfsize); // P(inbuf);
  fclose(protectedfile);

  // read eccfile and correct file
  int errors = 0, i;
  char *protbuf;                 // data from the ecc file
  if (verbose) {
    std::cout << "Reading " << eccfname << std::endl;
    std::cout.flush();
  }
  protbuf =   readandinterleave(eccblocks,ostripeskip,eccfile);
  // P(protbuf); P(crc(protbuf,eccblocks*blklen*sizeof(char))); 
  char *protptr=protbuf;
  if (verbose) {
    std::cout << "Calculating ... " << std::endl; std::cout.flush();
  }
  for(i=0; i<protfblocks ; i+=hblen) {
    bool correctable = hyperblockcorrect(inbuf+i*blklen,protptr);
    if (!correctable) std::cout << "uncorr(" << i << ") ";
    errors += correctable ? 0 : 1;
    protptr += hbcorr*blklen;
  }
  if (errors > 0) {
    std::cout << "WARNING: " << errors << " uncorrectable out of "
              << (protfblocks/hblen) << " hyperblocks"
              << " in file " << correctedfname << "\n";
  }
  fclose(eccfile);
  
  // write the corrected file
  FILE* correctedfile=fopen(correctedfname,"wb");
  if (NULL == correctedfile) { perror(correctedfname); abort(); }
  // ensure that required size is available
  fseek(correctedfile, protfsize-1, SEEK_SET); 
  fwrite("e",1,1,correctedfile); rewind(correctedfile);
  if (verbose) {
    std::cout << "Writing " << protectedfname << std::endl; std::cout.flush();
  }
  writeanddeinterleave(inbuf,protfblocks,stripeskip,correctedfile);
  free(inbuf); fclose(correctedfile);
  // we have to truncate it again, since we might have written too much
  // at the last block.
  if ( 0 != truncate(correctedfname, protfsize) ) {
    perror(correctedfname); abort();
  }
  
  // return value
  if ( 0 == errors ) {
    if ( verbose ) std::cout << "File " << correctedfname << " is OK.\n";
    return true;
  }
  return false; // uncorrectable errors.
}
